/*
** Module   : TemporaryBuffer.java
** Abstract : Specialized record buffer which supports temp table semantics
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description---------------------
** 001 ECF 20060125   @24356 Created initial version. Extends RecordBuffer
**                           to provide temp table semantics.
** 002 ECF 20060414   @25545 Added clear() method. Removes all records
**                           from a virtual temp table.
** 003 ECF 20060504   @26008 Added runtime support for the creation and
**                           dropping of secondary temp tables. These are
**                           tables created to support the extent fields
**                           of a primary temp table.
** 004 GES 20060517   @26201 Match new Finalizable interface requirement.
** 005 ECF 20060608   @27002 Use updated signature for Persistence.list().
** 006 ECF 20060621   @27508 Fixed temp table cleanup. Ensure that
**                           dropping a table or removing records as part
**                           of multiplex scope closure is performed
**                           within a committed database transaction. Also
**                           added a per-context reference count facility
**                           for temp tables.
** 007 ECF 20060627   @27594 Fixed temp table creation and buffer scope
**                           opening. Temp table creation must occur
**                           within a transaction. The superclass'
**                           implementation of openScope() is invoked from
**                           the overriding method.
** 008 ECF 20060713   @28047 Use new DBUtils class.
** 009 ECF 20060719   @28103 Use updated signature for Persistence.list().
** 010 ECF 20060914   @29596 Implemented master/slave buffers. These are
**                           linked to one another by a common multiplex
**                           ID.
** 011 ECF 20060919   @29699 Override the parent's implementations of
**                           processChange(), incrementDMOUseCount(),
**                           decrementDMOUseCount(), commit(), and
**                           rollback(). The first three disable the
**                           parent class' aggressive record eviction
**                           policy for temp tables, and the latter two
**                           disable subtransaction-level undo support for
**                           temp tables. NOTE: some level of undo support
**                           may need to be implemented eventually.
** 012 ECF 20061010   @30312 Added support for undoable temp tables. By
**                           default, temp tables are not undoable,
**                           however, this setting can be overridden by
**                           using a new variant of the static define()
**                           method.
** 013 ECF 20061019   @30545 Fixed NPE in copyAllRows(). Must not iterate
**                           if no source records are returned by the list
**                           query.
** 014 ECF 20061116   @31238 Fixed shared buffers and slave buffers.
**                           Tables were not being cleaned up properly in
**                           certain situations.
** 015 ECF 20061121   @31346 Fixed commit problem. Force ChangeBroker to
**                           report a pending flush after persisting or
**                           deleting records.
** 016 ECF 20061211   @31657 Make temp-tables undoable by default. The
**                           no-undo variant requires a separate option to
**                           be declared in Progress and will convert to
**                           special variant of the define() method, so
**                           undoability is the more appropriate default.
** 017 ECF 20070130   @32043 Various cleanup. Introduced methods for table
**                           counting by context. Fixed invocations of
**                           superclass methods to use new signatures.
** 018 ECF 20070202   @32072 Fixed regression in copyAllRows(). Recent
**                           changes to bulk copy implementation had
**                           broken this implementation.
** 019 ECF 20070326   @32612 Fixed associate() method and improved
**                           removeRecords() implementation. The former
**                           required a conditional flush before copying
**                           input records. The latter now uses a bulk
**                           delete instead of a looping delete.
** 020 ECF 20070329   @32649 Fixed shared temp-table support. Added static
**                           method useShared() to access a previously
**                           established, shared temp-table definition.
** 021 ECF 20070412   @32922 Override parent's setPinnedLockType() method.
**                           Does nothing, since locking is a no-op for
**                           temp tables.
** 022 ECF 20070415   @33012 Fixed define() variant which creates a slave
**                           buffer. In some cases, master buffer was not
**                           being determined correctly, resulting in an
**                           NPE.
** 023 ECF 20070416   @33030 Integrated user interrupt handling.
** 024 ECF 20070508   @33463 Fixed associate() to properly handle a
**                           validation exception. New requirement based
**                           on a change to RecordBuffer.flush().
** 025 ECF 20070601   @33919 Fixed memory leak. Buffers were not being
**                           removed from context's domains map when the
**                           buffer was no longer needed, which pinned the
**                           instances in context local memory. Changed
**                           implementation of domains map to be a weak
**                           hash map.
** 026 ECF 20070602   @33927 Implemented deferred table drop optimization.
**                           Temp tables created for the current database
**                           session are not dropped until the session is
**                           being closed. This prevents a condition where
**                           the database repeatedly creates and drops the
**                           same tables as a secondary program which
**                           creates TemporaryBuffers and allows them to
**                           go out of scope is invoked within a loop of
**                           another program.
** 027 EVL 20070609   @34005 Adding explicit import of the class
**                           org.hibernate.type.Type to eliminate conflict
**                           with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 028 ECF 20070706   @34405 Added deregisteredSessionListener() method.
**                           Required by change to SessionListener
**                           interface. Also replaced StringBuffer with
**                           StringBuilder.
** 029 ECF 20070715   @34856 Fixed removeRecords(). Release current buffer
**                           record after bulk delete.
** 030 ECF 20070827   @34994 Added bulk delete and optimized multiplexing.
**                           Override parent's implementation of delete()
**                           variant which accepts a where clause and
**                           query substitution parms to perform a bulk
**                           delete. Omit multiplex ID from where clauses
**                           when possible.
** 031 ECF 20070914   @35071 Fixed regression introduced by H030 (@34994).
**                           When removing records during the closure of a
**                           multiplex scope, we have to ensure only those
**                           records associated with the multiplex ID of
**                           the closing scope are removed from the temp
**                           table.
** 032 ECF 20070914   @35294 Fixed ConnectionManager notifications. We
**                           override the parent's implementation of two
**                           protected methods which notify the connection
**                           manager of buffer open and close events to do
**                           nothing, because we don't want to trace these
**                           for temp tables. Replaced database name
**                           string with database information object.
** 033 ECF 20071002   @35332 Exposed certain methods as public. Expanded
**                           use of generics.
** 034 ECF 20071018   @35592 Integrated PersistenceFactory.
** 035 ECF 20071108   @35909 Made nextMPID access threadsafe.
** 036 ECF 20071130   @36126 Integrated generics.
** 037 ECF 20071220   @36548 Modified Context.sessionClosing(). We now do
**                           more state cleanup when the session is
**                           closing due to an error.
** 038 SVL 20071231   @36628 Next primary key generation is carried out
**                           to the separate function which overrides the
**                           one in RecordBuffer.
** 039 ECF 20080102   @36638 Replaced use of deprecated nextPrimaryKey()
**                           method in Persistence. Primary key generation
**                           is now managed within Context inner class.
** 040 ECF 20080103   @36696 Fixed regression caused by #039 (@36638).
**                           Used wrong buffer to call nextPrimaryKey()
**                           from copyAllRows().
** 041 ECF 20080108   @36732 Fixed bulk delete. Changed call to delete()
**                           method in Persistence.
** 042 ECF 20080310   @37487 Fixed potential memory leak. copyAllRows()
**                           could leave DMOs in the Hibernate session if
**                           they were never loaded into a buffer. Also
**                           integrated DDL to drop indexes and to sweep
**                           index names in create and drop statements.
** 043 CA  20080430   @38161 Changes due to the TempTableHelper.sql*Drop
**                           /Create method signature change: instead of
**                           a Persistence object, the first parameter is
**                           a Database object.
** 044 SVL 20080516   @38292 Some superclass methods were overridden to
**                           no-ops to reduce memory consumption.
** 045 CA  20080530   @38458 Temporary tables passed as parameters to
**                           external procedures must match using each
**                           field's position and type, not its name;
**                           PROGRESS will not care if field names are
**                           different. Ref: #38457
** 046 ECF 20080602   @38615 Override getIndexedProperties(boolean) to do
**                           nothing. This method supports dirty checking,
**                           which is not needed for temp tables.
** 047 ECF 20080609   @38654 Backed out #046 (@38615) which regressed
**                           unique constraint validation. Removed
**                           getIndexProperties(boolean) method and added
**                           postValidateDirtyShare(), which overrides the
**                           parent's method to do nothing.
** 048 CA  20080605   @38570 The multiplex ID must start from 1.
**                           Implemented reversible bulk delete action
**                           for temporary buffers.
** 049 ECF 20080708   @39075 Fixed copyAllRows(). When copying from one
**                           table type to another, properties which are
**                           not available via a DMO's API must be
**                           skipped (i.e., computed columns).
** 050 ECF 20080723   @39172 Fixed flushing behavior for temp-tables with
**                           no indexes. Updates to such tables must be
**                           flushed to the database immediately.
** 051 ECF 20080806   @39329 Added rowid support. New primary keys are
**                           64-bit integers.
** 052 CA  20080808   @39360 For no-undo tables, the bulk delete will 
**                           always force removal of the records. Also,
**                           a reversible is registered with so the delete
**                           will be re-executed if db tx is rolled back.
** 053 ECF 20080815   @39558 Support global DMO change events.
** 054 CA  20080815   @39468 Support API change in DBUtils.
** 055 SIY 20080904   @39720 Inferred generics in buffer creation methods.
** 056 ECF 20080925   @39950 Refactored Reversible concrete classes.
** 057 ECF 20080926   @39954 Fixed ReversibleBulkDelete.setDMO(). Should
**                           not throw UnsupportedOperationException.
** 058 ECF 20081009   @40083 Integrated HQLPreprocessor API change.
** 059 ECF 20081022   @40212 Modified to adjust for Reversible API change.
** 060 ECF 20081127   @40710 Implement reclamation of temp table primary
**                           keys. On delete, a record's primary key is
**                           collected for possible recycling. On full
**                           transaction commit/rollback, pending keys are
**                           made available for re-use, or discarded,
**                           respectively. Currently, keys surrendered by
**                           bulk deletes and by rolled back record
**                           creations are not reclaimed.
** 061 ECF 20081128   @40718 Fixed regression introduced by #060 (@40710).
**                           Data structures for recycling primary keys
**                           were not being managed by table properly.
** 062 ECF 20081201   @40734 Modified primary key reclamation strategy.
**                           We now piggyback off the implementation in
**                           the superclass, so that surrendered keys are
**                           collected during the buffer's commit() and
**                           rollback() operations. This fixes yet another
**                           regression caused by the previous two mods.
** 063 ECF 20081202   @40744 Fixed regression in key reclamation. Must
**                           clear key reclamation data structures when
**                           the session is closed to avoid unique
**                           constraint violations in following session.
** 064 ECF 20081211   @40871 Improved logging in temp table creation code.
** 065 ECF 20090107   @41025 Fixed record release during bulk delete. A
**                           bulk delete event is now broadcast, which
**                           allows listening buffers to release their
**                           current records if they were deleted.
** 066 ECF 20090112   @41164 Fixed undo-related locking behavior. Locks
**                           were being downgraded rather than released in
**                           certain transaction-boundary cases. The
**                           change necessary in this class was to
**                           override certain lock-related methods from
**                           the superclass to provide no-op behavior.
** 067 ECF 20090218   @41334 Added safety code for dealing with maps of
**                           Reversibles. BufferManager methods which
**                           return these maps can now return null.
** 068 ECF 20090227   @41446 Attempted to address memory leak. Remove all
**                           Reversibles from BufferManager when the
**                           associated virtual table is no longer in use.
** 069 GES 20090424   @41943 Import change.
** 070 ECF 20090508   @42128 Added inner class TempValidationHelper. Extends
**                           RecordBuffer's implementation to override certain
**                           methods.
** 071 ECF 20090610   @42634 Optimized copyAllRows(). Call BufferManager's
**                           evictDMOIfUnused() method instead of incrementing
**                           then decrementing DMO use counts.
** 072 ECF 20090701   @43223 Various fixes and optimizations. Changed
**                           copyAllRows() method to use scrollable results
**                           instead of a results list. Added table empty
**                           tracking. Removed immediate flush concept.
**                           Refactored initialization to be lazy, upon open
**                           scope event.
** 073 ECF 20090717   @43230 Fixed regressions in #072 (@43223). Opening a
**                           multiplex scope was unconditionally setting the
**                           emptyTempTables flag for that multiplex ID, but
**                           it must only do so if the reference count is 1.
**                           When closing the last scope using a physical temp
**                           table, we must remove all the rows before the
**                           table is dropped, otherwise we can get into
**                           trouble flushing deferred rollbacks.
** 074 ECF 20090724   @43423 Added getMasterBuffer() method. This is needed to
**                           support proper rollback handling of shared temp
**                           tables.
** 075 ECF 20090729   @43450 Fixed copyAllRows(). Call to ChangeBroker to
**                           force a pending flush must come before the micro
**                           transaction commit, otherwise the copied records
**                           are lost. Also implemented no-op validate() call
**                           for Context, which implements Commitable.
** 076 ECF 20090810   @43591 Changed to accommodate SessionListener API
**                           change. sessionClosing() method was replaced with
**                           sessionEvent().
** 077 CA  20090916   @43916 Fixed bulk delete for no-undo temp tables. On
**                           commit, do not flush the bulk deletes for no-undo 
**                           temp tables, as this are relevant only for
**                           undoable tables.
** 078 CA  20091001   @44061 Fixed leak caused by bulk delete duplication: do 
**                           not duplicate it if the record was deleted 
**                           explicitly (via a DELETE operation) or implicitly
**                           (via a bulk delete).
** 079 CA  20091202   @44469 Reworked the rollback mechanism, to maintain 
**                           proper reversible order and thus to solve some
**                           unique-constraint violations during rollback of
**                           some bulk delete cases and mixed record operations.
**                           The ReversibleChain is no longer used, instead a 
**                           linked set with all operations which affect a 
**                           certain table is used.
** 080 OM  20130122          Added conversion support for CREATE TEMP-TABLE
**                           and related methods and attributes.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.InstantiationException;
import java.lang.reflect.*;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.*;
import org.hibernate.mapping.PersistentClass;
import org.hibernate.mapping.Property;
import org.hibernate.property.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * A specialized record buffer implementation which uses a temporary table
 * as backing storage and supports Progress' temp/work-table semantics.  A
 * new instance is created by invoking the static {@link #define} method,
 * which, like the superclass' method of the same name, creates a proxy
 * object which client code uses to interact with the database.
 * <p>
 * <b>Implementation Notes</b><br>
 * <b>Multiplexing.</b>  Because of limitations imposed by Hibernate's lack
 * of direct support for temp tables, a single temp table instance in the
 * backing database is multiplexed if multiple buffers of the same type of
 * DMO class are opened simultaneously in the same context.  This can happen
 * either when calling code explicitly defines multiple temp table buffers
 * using the same DMO type, or when method calls pass a temp table DMO proxy
 * as a parameter.  Multiplexing is used to enable Progress' semantic of
 * copying data between temp tables when they are passed as parameters to
 * nested procedure calls.
 * <p>
 * Multiplexing is achieved by adding a multiplex ID column to every temp
 * table schema.  Each buffer instance is assigned a unique multiplex ID the
 * first time a scope is opened on it.  Every access to the backing temp
 * table (i.e., queries, updates, inserts, deletes) take the multiplex ID
 * into account, such that the backing table is managed as multiple, virtual
 * table instances, one per distinct multiplex ID.
 * <p>
 * A temp table is created (along with its indexes, if any) the first time a
 * particular DMO type is accessed.  It is only dropped when no buffers
 * reference it.  As buffers go in and out of scope, a context local "in-use"
 * map is updated to reflect which buffers reference which physical temp
 * tables.  When a buffer goes out of scope, cleanup processing checks
 * whether it is the last buffer to reference its backing table.  If so, the
 * table is dropped.  If not, the table is not dropped, but all records with
 * that buffer's multiplex ID are removed from the table.
 * <p>
 * Temporary buffers may be scoped globally, or to the top-level (i.e.,
 * method) scope.  If global, cleanup is performed only when the context
 * itself is about to go out of scope.  If not global, cleanup is performed
 * when the method scope is popped by the transaction manager.  The latter
 * coincides with the external procedure scope in Progress.
 * <p>
 * There is a notion of a <i>master</i> buffer and <i>slave</i> buffers.  The
 * former is created as the result of converting a DEFINE TEMP-TABLE Progress
 * statement.  The latter is created as the result of converting a DEFINE
 * BUFFER FOR <i>temp-table</i> Progress statement.  A master buffer and all
 * of the associated slave buffers share the same multiplex ID, since these
 * buffers must all point to the same backing, virtual temp table for their
 * backing data.  The first time any buffer in such a group has a multiplex
 * ID assigned, all buffers in the group have their multiplex IDs assigned to
 * the same value.
 */
public final class TemporaryBuffer
extends RecordBuffer
{
   /** Name of the multiplex ID field in temp table DMO classes */
   static final String MULTIPLEX_FIELD_NAME = "_multiplex";
   
   /** Logger */
   private static final Log LOG = LogFactory.getLog(TemporaryBuffer.class);
   
   /** Multiplex ID sets currently in use for this context by DMO interface */
   private static final ContextLocal<Context> context =
   new ContextLocal<Context>()
   {
      protected Context initialValue() { return (new Context()); }
   };
   
   /** Hibernate parameter types passed to copyAllRows() query */
   private static final Type[] COPY_ALL_ROWS_QUERY_PARM_TYPES =
      DBUtils.makeTypeArray(new Object[] { new Integer(0) });
   
   /** Hibernate parameter types passed to bulk delete state change query */
   private static final Type[] BULK_DELETE_EVENT_QUERY_PARM_TYPES =
      DBUtils.makeTypeArray(new Object[] { new Long(0L) });
   
   /** Hibernate parameter types passed to bulk delete check query */
   private static final Type[] BULK_DELETE_CHECK_QUERY_PARM_TYPES =
      DBUtils.makeTypeArray(new Object[] { new Long(0L), new Integer(0) });
   
   /** Does this buffer represent a global temp table? */
   private final boolean global;
   
   /** Can property changes on this buffer be undone? */
   private final boolean undoable;
   
   /** Constructor used to create new instances of the backing DMO */
   private Constructor<?> dmoCtor = null;
   
   /** HQL used to check whether a record was deleted by a bulk delete */
   private String bulkDeleteEventHQL = null;
   
   /** HQL used to check whether a record was deleted from a virtual table */
   private String bulkDeleteCheckHQL = null;
   
   /** Destination DMO for output copy if necessary before scope closes */
   private Temporary outputDMO = null;
   
   /** Multiplex ID assigned to this buffer instance */
   private Integer multiplexID = null;
   
   /** Master buffer for this buffer;  null indicates this is the master */
   private TemporaryBuffer master = null;
   
   /**
    * Constructor. This class is only instantiated via the {@link #define}
    * factory method.  The implementation class associated with the given
    * <code>dmoIface</code> must have a constructor which accepts a multiplex
    * ID and must contain a <code>_multiplex</code> instance variable of type
    * <code>Integer</code>.
    *
    * @param   dmoIface
    *          Interface which defines DMO's API.  This interface MUST extend
    *          {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer is associated
    *          in the transaction manager whenever it is registered.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    * @param   undoable
    *          <code>true</code> if changes made to records managed by this
    *          buffer can be undone;  <code>false</code> if such changes are
    *          committed immediately, such that they cannot be undone.
    */
   private TemporaryBuffer(Class<?> dmoIface,
                           String variable,
                           boolean global,
                           boolean undoable)
   {
      super(dmoIface, DatabaseManager.TEMP_TABLE_DB.getName(), variable);
      
      this.global = global;
      this.undoable = undoable;
   }
   
   /**
    * Define a new buffer which is based upon the specified interface. The
    * buffer is permanently associated with the temp table database. It will
    * be registered with the transaction manager under the specified variable
    * name.
    * 
    * @param   <T> 
    *          Type of constructed instance.
    * @param   dmoIface
    *          Interface which defines DMO's public API.  This must be a
    *          sub-interface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    *
    * @return  An object which implements the specified interface, but which
    *          initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>dmoIface</code> is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Class<T> dmoIface,
                                                String variable,
                                                boolean global)
   {
      TemporaryBuffer buffer = makeBuffer(dmoIface, variable, global, true);
      
      Map<TemporaryBuffer, Object> domain =
         new WeakHashMap<TemporaryBuffer, Object>();
      domain.put(buffer, null);
      Map<TemporaryBuffer, Map<TemporaryBuffer, Object>> domains =
         context.get().domains;
      domains.put(buffer, domain);
      
      return createProxy(dmoIface, buffer);
   }
   
   /**
    * Define a new buffer which is based upon the specified interface. The
    * buffer is permanently associated with the temp table database. It will
    * be registered with the transaction manager under the specified variable
    * name.
    * 
    * @param   <T> 
    *          Type of constructed instance.
    * @param   dmoIface
    *          Interface which defines DMO's public API.  This must be a
    *          sub-interface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    * @param   undoable
    *          <code>true</code> if changes made to records managed by this
    *          buffer can be undone;  <code>false</code> if such changes are
    *          committed immediately, such that they cannot be undone.
    *
    * @return  An object which implements the specified interface, but which
    *          initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>dmoIface</code> is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Class<T> dmoIface,
                                                String variable,
                                                boolean global,
                                                boolean undoable)
   {
      TemporaryBuffer buffer = makeBuffer(dmoIface,
                                          variable,
                                          global,
                                          undoable);
      
      Map<TemporaryBuffer, Object> domain =
         new WeakHashMap<TemporaryBuffer, Object>();
      domain.put(buffer, null);
      Map<TemporaryBuffer, Map<TemporaryBuffer, Object>> domains =
         context.get().domains;
      domains.put(buffer, domain);
      
      return createProxy(dmoIface, buffer);
   }
   
   /**
    * Define a new buffer which is based upon a &quot;master&quot; buffer.
    * This is used as an analog to the DEFINE BUFFER statement when defining
    * a buffer for a temp-table or work-table.  The DMO interface and global
    * setting of the master buffer are used for the new buffer.  This buffer
    * will have the same multiplex ID as the master buffer, once it is
    * assigned to either buffer.
    *
    * @param   <T> 
    *          Type of constructed instance.
    * @param   template
    *          An instance of <code>Temporary</code> which is associated with
    *          a temporary buffer that serves as a template for the new
    *          buffer.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    *
    * @return  An object which implements the DMO interface of the master
    *          buffer, but which initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>master</code>'s DMO interface is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Temporary template, 
                                                String variable)
   {
      TemporaryBuffer m = (TemporaryBuffer) get(template);
      TemporaryBuffer masterBuf = (m.master == null ? m : m.master);
      Class<?> dmoIface = masterBuf.getDMOInterface();
      TemporaryBuffer buffer = makeBuffer(dmoIface,
                                          variable,
                                          masterBuf.global,
                                          masterBuf.undoable);
      buffer.master = masterBuf;
      buffer.multiplexID = masterBuf.multiplexID;
      
      Map<TemporaryBuffer, Map<TemporaryBuffer, Object>> domains =
         context.get().domains;
      Map<TemporaryBuffer, Object> domain = domains.get(masterBuf);
      domain.put(buffer, null);
      
      return (T) createProxy(dmoIface, buffer);
   }
   
   /**
    * Define a new buffer which is based upon a &quot;master&quot; buffer
    * stored in the {@link com.goldencode.p2j.util.SharedVariableManager
    * SharedVariableManager}.  This is used as an analog to the DEFINE SHARED
    * TEMP-TABLE statement when defining a buffer for a shared temp-table or
    * work-table which has been established elsewhere.  The DMO interface,
    * global, and undoability settings of the master buffer are used for the
    * new buffer.  This buffer will have the same multiplex ID as the master
    * buffer, which should have been assigned by an enclosing scope by the
    * time this method is invoked.
    *
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager and shared variable
    *          manager.
    *
    * @return  An object which implements the DMO interface of the master
    *          buffer stored in the shared variable manager, but which
    *          initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>master</code>'s DMO interface is not assignable from
    *          <code>Temporary</code>.
    */
   public static Temporary useShared(String variable)
   {
      Temporary master = SharedVariableManager.lookupTempTable(variable);
         
      return define(master, variable);
   }
   
   /**
    * Associate two temp tables with one another, such that all relevant
    * records can be copied between them at prescribed times.  If an
    * <code>input</code> copy is performed, all records are first removed
    * from the destination table.
    *
    * @param   srcDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the source temporary buffer.
    * @param   dstDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the destination temporary buffer.
    * @param   input
    *          if <code>true</code>, records are copied from source table to
    *          destination table immediately upon invocation of this method,
    *          otherwise they are not.
    * @param   output
    *          if <code>true</code>, records are copied from destination
    *          table back to source table when the current transaction or
    *          subtransaction is committed;  otherwise they are not.
    *
    * @throws  ErrorConditionException
    *          if a recoverable error occurs copying records.
    * @throws  RuntimeException
    *          if an unrecoverable error occurs invoking methods on the
    *          underlying DMOs to get or set data values.
    */
   public static void associate(Temporary srcDMO,
                                Temporary dstDMO,
                                boolean input,
                                boolean output)
   {
      associate(srcDMO, dstDMO, input, output, false);
   }
   
   /**
    * Associate two temp tables with one another, such that all relevant
    * records can be copied between them at prescribed times.
    *
    * @param   srcDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the source temporary buffer.
    * @param   dstDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the destination temporary buffer.
    * @param   input
    *          if <code>true</code>, records are copied from source table to
    *          destination table immediately upon invocation of this method,
    *          otherwise they are not.
    * @param   output
    *          if <code>true</code>, records are copied from destination
    *          table back to source table when the current transaction or
    *          subtransaction is committed;  otherwise they are not.
    * @param   append
    *          <code>true</code> to add new records without affecting
    *          existing records;  <code>false</code> to remove all existing
    *          records before the copy takes place.  This option affects the
    *          <code>input</code> copy only, not the <code>output</code> copy.
    *
    * @throws  ErrorConditionException
    *          if a recoverable error occurs flushing the source buffer in
    *          preparation for an input copy or while copying records.
    * @throws  RuntimeException
    *          if an unrecoverable error occurs invoking methods on the
    *          underlying DMOs to get or set data values.
    */
   public static void associate(final Temporary srcDMO,
                                final Temporary dstDMO,
                                boolean input,
                                boolean output,
                                boolean append)
   {
      if (input)
      {
         TemporaryBuffer srcBuf =
            (TemporaryBuffer) ((BufferReference) srcDMO).buffer();
         Persistence p = srcBuf.getPersistence();
         try
         {
            boolean shouldCommit = srcBuf.isTransient() &&
                                   p.beginTransaction();
            srcBuf.flush();
            if (shouldCommit)
            {
               p.commit();
            }
         }
         catch (ValidationException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
         
         copyAllRows(srcDMO, dstDMO, append);
      }
      
      // Set the destination DMO if output copy is required.  This is
      // actually the source DMO passed to this method, since the direction
      // of the copy is reversed on output.
      TemporaryBuffer dstBuf =
         (TemporaryBuffer) ((BufferReference) dstDMO).buffer();
      dstBuf.outputDMO = (output ? srcDMO : null);
   }
   
   /**
    * Remove all records from a (virtual) temp table.
    *
    * @param   dmo
    *          DMO instance returned by a previous call to {@link #define}.
    */
   public static void clear(Temporary dmo)
   {
      TemporaryBuffer buffer = (TemporaryBuffer) get(dmo);
      try
      {
         buffer.removeRecords(null, null, false);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Get an iterator on the names of all temp tables which currently exist
    * in the current context.
    * 
    * @return  Iterator on active table names.
    */
   public static Iterator<String> activeTables()
   {
      return context.get().activeTables.iterator();
   }
   
   /**
    * Report the number of temp tables which currently exist in the current
    * context.
    *
    * @return  Number of temp tables which have been created for the current
    *          context and not yet dropped.
    */
   public static int getTableCount()
   {
      return context.get().getTableCount();
   }
   
   /**
    * Factory method to construct a new temporary buffer.
    *
    * @param   dmoIface
    *          Interface which defines DMO's public API.  This must be a
    *          subinterface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    * @param   undoable
    *          <code>true</code> if changes made to records managed by this
    *          buffer can be undone;  <code>false</code> if such changes are
    *          committed immediately, such that they cannot be undone.
    *
    * @return  A new instance of this class.
    */
   private static TemporaryBuffer makeBuffer(Class<?> dmoIface,
                                             String variable,
                                             boolean global,
                                             boolean undoable)
   {
      if (!DatabaseManager.isTempTable(dmoIface))
      {
         throw new IllegalArgumentException(
            "DMO interface must extend Temporary");
      }
      
      return (new TemporaryBuffer(dmoIface, variable, global, undoable));
   }
   
   /**
    * Get a context-local <code>Commitable</code> which needs to have its
    * methods invoked on commit/rollback.
    * 
    * @return  The <code>Commitable</code> object for the current context.
    */
   static Commitable getContextCommitable()
   {
      return context.get();
   }
   
   /**
    * Copy all records from one temp table to another, optionally appending
    * to records already present in the destination table.
    * <p>
    * TODO:  it seems this method can be optimized to use a DMO's deepCopy()
    * method instead of reflection when the source and destination DMO types
    * are the same.
    *
    * @param   srcDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the source temporary buffer.
    * @param   dstDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the destination temporary buffer.
    * @param   append
    *          <code>true</code> to add new records without affecting
    *          existing records;  <code>false</code> to remove all existing
    *          records before the copy takes place.
    *
    * @throws  ErrorConditionException
    *          if a recoverable error occurs copying records.
    * @throws  RuntimeException
    *          if an unrecoverable error occurs invoking methods on the
    *          underlying DMOs to get or set data values.
    */
   @SuppressWarnings("unchecked")
   private static void copyAllRows(Temporary srcDMO,
                                   Temporary dstDMO,
                                   boolean append)
   {
      // Get source and destination buffers, both of which must be
      // TemporaryBuffer instances.
      TemporaryBuffer srcBuf =
         (TemporaryBuffer) ((BufferReference) srcDMO).buffer();
      TemporaryBuffer dstBuf =
         (TemporaryBuffer) ((BufferReference) dstDMO).buffer();
      
      srcBuf.initialize();
      dstBuf.initialize();
      
      Class<?> srcClass = srcBuf.getDMOImplementationClass();
      
      // Get maps of source getter methods and destination setter methods,
      // each keyed by property name.
      Map<String, Method> getters = srcBuf.getGetterMap();
      Map<String, Method> sizers = srcBuf.getSizerMap();
      Map<String, Method> setters = dstBuf.getSetterMap();
      
      // PROGRESS does not care if the field names are different; what it
      // cares about are only the field type and position - if a source field
      // has the same position and type as the destination field, then they
      // are a match. If the source and destination fields at a certain
      // position do not match, then PROGRESS will throw an error and abend
      // the application.
      
      // if the backing tables are different, check if the fields match
      if (srcBuf.getTable().equals(dstBuf.getTable()) == false)
      {
         Class<?> dstClass = dstBuf.getDMOImplementationClass();
         String srcBufName = srcBuf.getDMOAlias();
         String dstBufName = dstBuf.getDMOAlias();
         
         // resolve the tables for both DMOs
         PersistentClass srcTable = DatabaseManager
            .getClassMapping(DatabaseManager.TEMP_TABLE_DB, 
                             srcBuf.getDMOImplementationClass());
         PersistentClass dstTable = DatabaseManager
            .getClassMapping(DatabaseManager.TEMP_TABLE_DB, 
                             dstBuf.getDMOImplementationClass());
         
         // we need to modify the setter properties, so instantiate a new map
         setters = new HashMap<String, Method>(dstBuf.getSetterMap());
         
         // synchronize the dest property names with the source property names
         Iterator<?> srcIter = srcTable.getPropertyIterator();
         Iterator<?> dstIter = dstTable.getPropertyIterator();
         
         while (true)
         {
            Property srcProp = null;
            boolean srcHasNext = srcIter.hasNext();
            while (srcHasNext)
            {
               srcProp = (Property) srcIter.next();
               
               // Skip properties that are not accessible via the DMO's API.
               PropertyAccessor pa = srcProp.getPropertyAccessor(srcClass);
               if (pa instanceof NopPropertyAccessor)
               {
                  srcHasNext = srcIter.hasNext();
                  continue;
               }
               break;
            }
            
            Property dstProp = null;
            boolean dstHasNext = dstIter.hasNext();
            while (dstHasNext)
            {
               dstProp = (Property) dstIter.next();
               
               // Skip properties that are not accessible via the DMO's API.
               PropertyAccessor pa = dstProp.getPropertyAccessor(dstClass);
               if (pa instanceof NopPropertyAccessor)
               {
                  dstHasNext = dstIter.hasNext();
                  continue;
               }
               break;
            }
            
            if (!srcHasNext && !dstHasNext)
            {
               break;
            }
            
            // if the (accessible) property count is different, abend
            if (srcHasNext != dstHasNext)
            {
               ErrorManager.showErrorAndAbend(5363,
                  "The caller's temp-table parameter " + srcBufName +
                  " does not match to the target temp-table " + dstBufName);
            }
            
            String srcType = srcProp.getValue().getType().getName();
            String dstType = dstProp.getValue().getType().getName();
            
            if (srcType == null ||
                dstType == null ||
                !srcType.equals(dstType))
            {
               ErrorManager.showErrorAndAbend(5363,
                  "The caller's temp-table parameter " + srcBufName +
                  " does not match to the target temp-table " + dstBufName);
            }
            
            String srcColName = srcProp.getName();
            String dstColName = dstProp.getName();
            
            // synchronize the property names of the destination DMO with the 
            // property names of the source DMO
            if (srcColName.equals(dstColName) == false)
            {
               Method setter = setters.remove(dstColName);
               if (setter != null)
               {
                  setters.put(srcColName, setter);
               }
            }
         }
      }
      
      // Establish the set of properties being copied. This set will contain
      // ALL source getter properties, as the setters and getters maps were
      // synced.
      Set<String> props = getters.keySet();
      
      Persistence persistence = srcBuf.getPersistence();
      boolean savedAny = false;
      int commit = 0;   // 0 = do nothing, 1 = commit, -1 = rollback
      try
      {
         commit = (persistence.beginTransaction() ? 1 : 0);
         
         // If not in append mode, first remove records (if any) from
         // destination table.
         if (!append)
         {
            dstBuf.removeRecords(null, null, true);
         }
         
         // Get the list of source records to be copied.  These will be all
         // records in srcBuf's backing temp table which have srcDMO's
         // multiplex ID.
         StringBuilder buf = new StringBuilder("from ");
         buf.append(srcBuf.getDMOImplementationName());
         if (srcBuf.isMultiplexed())
         {
            buf.append(" where ");
            buf.append(MULTIPLEX_FIELD_NAME);
            buf.append(" = ?");
         }
         // force the records to be ordered by their ID
         buf.append(" order by id");
         
         String hql = buf.toString();
         Object[] parms = new Object[] { srcBuf.getMultiplexID() };
         ScrollableResults results =
            persistence.scroll(hql,
                               parms,
                               COPY_ALL_ROWS_QUERY_PARM_TYPES,
                               0,
                               0,
                               ScrollMode.FORWARD_ONLY);
         
         BufferManager bufMgr = BufferManager.get();
         
         while (results.next())
         {
            // Copy each source record into a new destination DMO instance.
            Persistable srcRec = (Persistable) results.get(0);
            Persistable dstRec = dstBuf.instantiateDMO();
            Serializable id = dstBuf.nextPrimaryKey();
            copy(srcRec,
                 (DataModelObject) dstRec,
                 getters,
                 setters,
                 sizers,
                 props);
            bufMgr.evictDMOIfUnused(persistence, srcRec);
            if (LOG.isTraceEnabled())
            {
               String msg =
                  "Copy temp table record from MPID " +
                  srcBuf.getMultiplexID() + " to MPID " +
                  dstBuf.getMultiplexID() + "...";
               LOG.trace(msg);
               LOG.trace(srcBuf.describe() + srcBuf.toString(srcRec));
               LOG.trace(dstBuf.describe() + dstBuf.toString(dstRec));
            }
            persistence.save(dstRec, id);
            savedAny = true;
         }
      }
      catch (PersistenceException exc)
      {
         commit *= -1;
         DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
         ErrorManager.recordOrThrowError(exc);
      }
      catch (IllegalAccessException exc)
      {
         commit *= -1;
         
         throw new RuntimeException("Error creating record", exc);
      }
      catch (InstantiationException exc)
      {
         commit *= -1;
         
         throw new RuntimeException("Error creating record", exc);
      }
      catch (InvocationTargetException exc)
      {
         commit *= -1;
         DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
         
         throw new RuntimeException("Error creating record", exc);
      }
      finally
      {
         if (savedAny)
         {
            ChangeBroker.get().forcePendingFlush(dstBuf);
         }
         
         try
         {
            switch (commit)
            {
               case 1:
                  persistence.commit();
                  break;
               case -1:
                  persistence.rollback();
                  break;
               case 0:
                  break;
            }
         }
         catch (PersistenceException exc)
         {
            DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
            ErrorManager.recordOrThrowError(exc);
         }
      }
   }
   
   /**
    * Restores this buffer's state to that which it had at the beginning of
    * the current block.  In particular, this rolls back all the reversible
    * items for the current scope and removes these entries from the rollback
    * dictionaries managed by the <code>BufferManager</code>.
    * <p>
    * Auto-commit buffers are never rolled back.
    *
    * @param   transaction
    *          <code>true</code> if this commit represents the master
    *          transaction rollback;  <code>false</code> if it represents a
    *          sub-transaction rollback.
    */
   public void rollback(boolean transaction)
   {
      if (!isUndoable())
      {
         // in case of the no-undo tables, we need to re-apply all the changes 
         // recorded during this transaction; so, the empty flag needs to be 
         // reset
         // TODO: maybe a restoration mechanism would be better ?
         Context local = context.get();
         local.emptyTableFlags.put(multiplexID, Boolean.FALSE);
      }
      
      super.rollback(transaction);
   }

   /**
    * Commit the changes made within a subtransaction.  This entails:
    * <ul>
    *   <li>copying all reversible, record-level actions associated with the
    *       current scope to the next higher scope, so rollback capability
    *       for these actions is not lost when the scope is popped;
    *   <li>copying all undoable, property-level changes associated with the
    *       current scope to the next higher scope, so rollback capability
    *       for these actions is not lost when the scope is popped;
    * </ul>
    * <p>
    * Nothing is done if this is a full transaction commit, because no
    * rollback will be possible at the parent scope once the database-level
    * transaction is committed.
    * <p>
    * If this is a full transaction and there are pending bulk actions, force
    * those actions to be executed (i.e. force the delete).
    *
    * @param   transaction
    *          <code>true</code> if this commit represents the master
    *          transaction commit;  <code>false</code> if it represents a
    *          sub-transaction commit.
    */
   public void commit(boolean transaction)
   {
      List<PersistenceException> errors =
         new ArrayList<PersistenceException>();

      if (transaction && isUndoable())
      {
         // flush the bulk deleted records which had their multiplex ID
         // changed; this is performed only for undoable tables.
         Set<Reversible> r0 = bufferManager.getReversibles(this, 0, false);
         if (r0 != null && !r0.isEmpty())
         {
            ReversibleBulkDelete rev = null;
            try 
            {
               Iterator<Reversible> iter = r0.iterator();
               while (iter.hasNext())
               {
                  Reversible r = iter.next();
                  
                  if (r instanceof ReversibleBulkDelete)
                  {
                     rev = (ReversibleBulkDelete) r;
                     rev.forceDelete();
                     iter.remove();
                  }
               }
            }
            catch (PersistenceException exc)
            {
               if (LOG.isErrorEnabled())
               {
                  LOG.error(describe() +
                            ":  error rolling back undoable bulk delete " +
                            rev,
                            exc);
               }
               
               errors.add(exc);
            }
         }
      }
      super.commit(transaction);

      // Report or eat errors.
      Iterator<PersistenceException> iter = errors.iterator();
      while (iter.hasNext())
      {
         PersistenceException exc = iter.next();
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Respond to a record change event.  Override the parent's implementation
    * to handle bulk delete events.  If our current record no longer exists in
    * the database, we must release it.
    *
    * @param   event
    *          Event which describes the DMO state change.
    *
    * @throws  PersistenceException
    *          if an error occurs executing a query to detect whether current
    *          record still exists.
    */
   public void stateChanged(RecordChangeEvent event)
   throws PersistenceException
   {
      if (event.isInsert())
      {
         Context local = context.get();
         local.emptyTableFlags.put(multiplexID, Boolean.FALSE);
         
         super.stateChanged(event);
      }
      if (event.isBulkDelete())
      {
         if (!isActive())
         {
            return;
         }
         
         Persistable dmo = getCurrentRecord();
         if (dmo == null)
         {
            // No release necessary.
            return;
         }
         
         // We can avoid issuing a query if the event represents a full bulk
         // delete and the deleted multiplex ID matches our own, because we
         // know our record must have been deleted.
         if (event.isFullBulkDelete())
         {
            Integer srcMPID = event.getSource().getMultiplexID();
            if (srcMPID.equals(multiplexID))
            {
               release();
               
               return;
            }
         }
         
         // Query whether the buffer's current record still exists.
         Serializable id = dmo.getId();
         Persistence persistence = getPersistence();
         if (bulkDeleteEventHQL == null)
         {
            String entity = getEntityName();
            StringBuilder buf = new StringBuilder();
            buf.append("select tt.id from ");
            buf.append(entity);
            buf.append(" tt where tt.id=?");
            bulkDeleteEventHQL = buf.toString();
         }
         List list = persistence.list(bulkDeleteEventHQL,
                                      new Object[] { id },
                                      BULK_DELETE_EVENT_QUERY_PARM_TYPES,
                                      1,
                                      0,
                                      true);
         if (list == null)
         {
            release();
         }
      }
      else
      {
         super.stateChanged(event);
      }
   }
   
   /**
    * Initialize the core data on which this buffer relies.  This should be
    * called only once per buffer instance, when the buffer's scope is first
    * opened.
    * 
    * @return  <code>true</code> if the buffer was initialized with this call;
    *          <code>false</code> if the buffer had been initialized with a
    *          previous call to this method.
    *
    * @throws  IllegalArgumentException
    *          if no implementation class is found for the given database and
    *          interface.
    * @throws  RuntimeException
    *          if the backing DMO implementation class does not implement a
    *          constructor which accepts a multiplex ID.
    */
   protected boolean initialize()
   {
      if (!super.initialize())
      {
         return false;
      }
      
      Class<?> dmoCls = getDMOImplementationClass();
      try
      {
         dmoCtor = dmoCls.getDeclaredConstructor(new Class<?>[]
                                                 { Integer.class });
         Field fld = dmoCls.getDeclaredField(MULTIPLEX_FIELD_NAME);
         if (!fld.getType().equals(Integer.class))
         {
            String msg = "Invalid temporary table DMO:  " +
                         dmoCls.getName() +
                         "." +
                         MULTIPLEX_FIELD_NAME +
                         " field must be of type Integer";
            
            throw new RuntimeException(msg);
         }
      }
      catch (NoSuchMethodException exc)
      {
         String msg = "Invalid temporary table DMO:  " + dmoCls.getName() +
            " does not implement a multiplex constructor";
            
         if (LOG.isErrorEnabled())
         {
            LOG.error(describe() + ":  " + msg, exc);
         }
         
         throw new RuntimeException(msg, exc);
      }
      catch (NoSuchFieldException exc)
      {
         String msg = "Invalid temporary table DMO:  " +
                      dmoCls.getName() +
                      " does not declare a " +
                      MULTIPLEX_FIELD_NAME +
                      " field";
            
         if (LOG.isErrorEnabled())
         {
            LOG.error(describe() + ":  " + msg, exc);
         }
         
         throw new RuntimeException(msg, exc);
      }
      
      return true;
   }
   
   /**
    * Get the logical database name associated with this buffer.  This will
    * always be the standard name for the temp table database as defined in
    * {@link DatabaseManager#TEMP_TABLE_DB}, regardless of the value of the
    * <code>ldbOrAlias</code> parameter.
    *
    * @param   ldbOrAlias
    *          A logical database name or database alias.
    *
    * @return  Standard, temp table database name.
    */
   protected String getLDBName(String ldbOrAlias)
   {
      return DatabaseManager.TEMP_TABLE_DB.getName();
   }
   
   /**
    * Get the database information object associated with this buffer.  This
    * will always be the standard temp table database as defined in
    * {@link DatabaseManager#TEMP_TABLE_DB}, regardless of the value of the
    * <code>ldbName</code> parameter.
    *
    * @param   ldbName
    *          A logical database name.
    *
    * @return  Standard, temp table database.
    */
   protected Database getPDB(String ldbName)
   {
      return DatabaseManager.TEMP_TABLE_DB;
   }
   
   /**
    * Register this buffer with the connection manager.
    * <p>
    * This implementation overrides the parent method to do nothing, since
    * accesses to the temp table database are outside the scope of the
    * connection manager's responsibilities.
    */
   protected void openConnectionManagerScope()
   {
   }
   
   /**
    * Deregister this buffer with the connection manager.
    * <p>
    * This implementation overrides the parent method to do nothing, since
    * accesses to the temp table database are outside the scope of the
    * connection manager's responsibilities.
    */
   protected void closeConnectionManagerScope()
   {
   }
   
   /**
    * Create a new instance of the DMO implementation class associated with
    * this buffer, and assign this buffer's multiplex ID to it.  This method
    * must not be invoked before a buffer scope has been opened on this
    * buffer.
    *
    * @return  A transient DMO instance.
    *
    * @throws  InstantiationException
    *          if the class cannot be instantiated.
    * @throws  IllegalAccessException
    *          if the DMO class' target constructor cannot be accessed.
    * @throws  InvocationTargetException
    *          if the underlying constructor throws an exception.
    */
   protected Persistable instantiateDMO()
   throws InstantiationException,
          IllegalAccessException,
          InvocationTargetException
   {
      if (multiplexID == null)
      {
         throw new IllegalStateException(
            "Cannot create a new record before buffer scope is opened");
      }
      
      return (Persistable) dmoCtor.newInstance(multiplexID);
   }
   
   /**
    * Convenience method to indicate whether this buffer is backed by a
    * temporary table.
    * 
    * @return  <code>true</code>.
    */
   protected boolean isTemporary()
   {
      return true;
   }
   
   /**
    * Indicate whether the table backing this buffer is known to contain no
    * rows.
    * <p>
    * This implementation involves only the virtual table backing the current
    * buffer (i.e., whose multiplex ID matches that of this buffer), not the
    * entire physical table.
    * 
    * @return  <code>true</code> if the table has no rows;  <code>false</code>
    *          if the table might have rows.
    */
   protected boolean isTableDefinitelyEmpty()
   {
      Context local = context.get();
      Boolean empty = local.emptyTableFlags.get(multiplexID);
      
      return (empty != null && empty.booleanValue());
   }
   
   /**
    * Indicate whether changes made to records managed by this buffer can be
    * undone during a rollback.
    *
    * @return  <code>true</code> if record changes can be undone, else
    *          <code>false</code>.
    */
   protected boolean isUndoable()
   {
      return undoable;
   }
   
   /**
    * Override the superclass method to do nothing.  Locking is a no-op for
    * temp table records, since each user has exclusive access to their temp
    * table instance.
    *
    * @throws  PersistenceException
    *          never.
    */
   protected void upgradeLock()
   throws PersistenceException
   {
   }
   
   /**
    * Override the parent's implementation to do nothing.
    */
   protected void pushPinnedLockScope()
   {
   }
   
   /**
    * Override the parent's implementation to do nothing.
    * 
    * @param   transaction
    *          <code>true</code> if we are at a full transaction boundary,
    *          else <code>false</code>.
    * @param   undo
    *          <code>true</code> if we are processing an undo on the current
    *          scope, else <code>false</code>.
    */
   protected void popPinnedLockScope(boolean transaction, boolean undo)
   {
   }
   
   /**
    * Get the multiplex ID of this temporary buffer.  This will be
    * <code>null</code> if this method is invoked before a buffer scope is
    * opened on this buffer for the first time.  Otherwise, it will be a
    * unique identifier within the current context.
    *
    * @return  Multiplex ID for this temporary buffer.
    */
   protected Integer getMultiplexID()
   {
      return multiplexID;
   }
   
   /**
    * Indicate whether the backing table for this buffer is multiplexed;  that
    * is, whether the backing table represents multiple, virtual tables by
    * using a multiplex ID to differentiate table segments.
    * <p>
    * Temp tables always are multiplexed.
    *  
    * @return  <code>true</code>.
    */
   protected boolean isMultiplexed()
   {
      return true;
   }
   
   /**
    * Indicate whether this buffer implementation allows bulk actions.
    *  
    * @return  <code>true</code>.
    */
   protected boolean allowsBulkActions()
   {
      return true;
   }

   /**
    * Retrieve the master record buffer for this buffer.
    * <p>
    * This only has meaning for shared temp tables.  For a temp table defined
    * as NEW SHARED, this buffer is returned.  For a temp table defined as
    * SHARED, the domain master buffer is returned.
    * 
    * @return  The master buffer.
    */
   protected RecordBuffer getMasterBuffer()
   {
      return (master != null ? master : this);
   }
   
   /**
    * Open a new buffer scope for this record buffer.  Open a new multiplex
    * scope and assign a multiplex ID the first time we are invoked for a
    * buffer instance.
    */
   protected void openScope()
   {
      super.openScope();
      
      // Only need to process if we are opening the outermost scope for this
      // buffer.  Nested scopes need not apply.
      if (getOpenScopeCount() == 1)
      {
         // A temporary buffer can be initialized as soon as its first scope
         // is opened, because the _temp database is permanently connected.
         initialize();
         
         // If this is the first time this table is being used, create it in
         // the database.
         if (openMultiplexScope())
         {
            context.get().createTable(getDMOInterface(), getTable());
         }
         
         // Register for temp table cleanup.
         Multiplexer multiplexer = new Multiplexer();
         if (global)
         {
            // A global temp table buffer is cleaned up when the context ends.
            TransactionManager.registerFinalizable(multiplexer, true);
         }
         else
         {
            // A non-global temp table buffer is cleaned up when the top
            // level method scope ends (this corresponds with the external
            // procedure in Progress).
            TransactionManager.registerTopLevelFinalizable(multiplexer, true);
         }
      }
   }
   
   /**
    * Initialize the validation helper object.
    * 
    * @return  Validation helper object.
    */
   protected ValidationHelper initValidationHelper()
   {
      return (new TempValidationHelper());
   }
   
   /**
    * Override the superclass method to always return LockType.NONE. Locking
    * is a no-op for temp table records, since each user has exclusive access
    * to their temp table instance.
    *
    * @param  id
    *         Not used.
    *
    * @return LockType.NONE
    */
   protected LockType getPinnedLockType(Serializable id)
   {
      return LockType.NONE;
   }
   
   /**
    * Override the superclass method to do nothing.  Locking is a no-op for
    * temp table records, since each user has exclusive access to their temp
    * table instance.
    * 
    * @param   id
    *          Not used.
    * @param   pinnedLockType
    *          Not used.
    */
   protected void setPinnedLockType(Serializable id,
                                    LockType pinnedLockType)
   {
   }
   
   /**
    * Override the superclass method to do nothing.  Locking is a no-op for
    * temp table records, since each user has exclusive access to their temp
    * table instance.
    */
   protected void initPinnedLockTypes()
   {
   }
   
   /**
    * Create and set the current bulk delete reversible action.
    * This variant produces a <code>ReversibleBulkDelete</code> instance.
    * <p>
    * If this block is not undoable, this method is a no-op.
    * 
    * @param   mpid
    *          The multiplex ID used to mark those records as deleted.
    * @param   hasWhere
    *          Flag indicating if this bulk delete had a where clause or not.
    */
   private void setReversibleBulkDelete(int mpid, boolean hasWhere)
   {
      // If we are in a transaction, remember this record change so it can be
      // undone at rollback if necessary.
      if (TransactionManager.isTransaction())
      {
         ReversibleBulkDelete item = new ReversibleBulkDelete(mpid, hasWhere);
         
         bufferManager.addReversible(this, null, item, false);

         if (LOG.isDebugEnabled())
         {
            LOG.debug(describe() + ":  adding reversible [" + item + "]");
         }
      }
   }

   /**
    * Open a new multiplex scope for this buffer.  The return value indicates
    * whether it is the first time a multiplex scope has been opened for this
    * buffer's DMO interface in this context.  If so, the caller will need to
    * create the backing temp table in the database.  This method permanently
    * assigns the multiplex ID to this buffer instance.
    *
    * @return  <code>true</code> if this is the first multiplex scope opened
    *          for the underlying DMO type;  <code>false</code> if multiplex
    *          scopes are already open in this context for this DMO type.
    */
   private boolean openMultiplexScope()
   {
      Class<?> dmoIface = getDMOInterface();
      Context local = context.get();
      Map<Class<?>, Map<Integer, Integer>> map = local.inUseMPIDs;
      
      // Get the map of multiplex IDs which are in use for this DMO type
      // within our context.  The value for each multiplex ID key is the
      // reference count of buffers using that ID.
      Map<Integer, Integer> inUse = map.get(dmoIface);
      
      if (inUse == null)
      {
         // This is the first use of this DMO type for this context;  create
         // a new in-use map.
         inUse = new HashMap<Integer, Integer>();
         map.put(dmoIface, inUse);
      }
      
      boolean createTable = inUse.isEmpty();
      
      int refCount = 0;
      
      if (multiplexID == null)
      {
         // Generate a unique multiplex ID.
         Integer id = local.getNextMPID(inUse);
         
         // Assign the unique ID as the permanent multiplex ID of every
         // buffer in this buffer's domain.
         Object key = getMasterBuffer();
         Map<TemporaryBuffer, Object> domain = local.domains.get(key);
         Iterator<TemporaryBuffer> iter = domain.keySet().iterator();
         while (iter.hasNext())
         {
            TemporaryBuffer next = iter.next();
            next.multiplexID = id;
         }
         
         // Start off the reference count.
         refCount = 1;
      }
      else
      {
         // Increment the existing reference count for this multiplex ID.
         refCount = inUse.get(multiplexID) + 1;
      }
      
      inUse.put(multiplexID, refCount);
      
      // If this is the first use of this multiplex ID, the virtual table
      // must be empty.
      if (refCount == 1)
      {
         local.emptyTableFlags.put(multiplexID, Boolean.TRUE);
      }
      
      return createTable;
   }
   
   /**
    * Perform a bulk delete, restricted by the given where clause and query
    * substitution parameters.
    * <p>
    * The targeted records are not actually removed from the buffer; instead,
    * they are assigned another multiplex ID, which marks them as deleted.
    * <p>
    * Note:  primary keys surrendered via bulk deletes are not currently
    * recycled.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  PersistenceException
    *          if an error occurs performing the bulk delete operation.
    */
   protected void delete(String where, Object[] args)
   throws PersistenceException
   {
      removeRecords(where, args, false);
   }
   
   /**
    * Delete the current record from the database.  Set the current record to
    * null if the deletion is successful.  Reclaim the record's primary key
    * for later re-use.
    *
    * @throws  PersistenceException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   protected void delete()
   throws PersistenceException
   {
      boolean autoCommit = !TransactionManager.isTransaction();
      
      Persistence persistence = null;
      Serializable id = null;
      if (autoCommit)
      {
         persistence = getPersistence();
         persistence.beginTransaction();
         id = getCurrentRecord().getId();
      }
      
      super.delete();
      
      if (autoCommit)
      {
         Context local = context.get();
         local.reclaimKey(getDMOInterface(), id);
         persistence.commit();
         local.commit(true);
      }
   }
   
   /**
    * Reclaim the given list of primary keys for re-use.
    * 
    * @param   keys
    *          A list of primary keys which are available for re-use.  Must
    *          not be <code>null</code>.
    */
   protected void reclaimKeys(List<Serializable> keys)
   {
      Context local = context.get();
      Class<?> dmoIface = getDMOInterface();
      for (Serializable key : keys)
      {
         local.reclaimKey(dmoIface, key);
      }
   }

   /**
    * Returns the primary key for a new DMO associated with this buffer.
    *
    * @return  The primary key for a new DMO associated with this buffer.
    *
    * @throws  PersistenceException
    *          if there was an error while retrieving the next primary key.
    */
   protected Serializable nextPrimaryKey()
   throws PersistenceException
   {
      Context local = context.get();
      
      return local.nextPrimaryKey(getDMOInterface());
   }
   
   /**
    * Close the current multiplex scope for the underlying DMO type in the
    * current context.  Remove all records associated with this buffer's
    * multiplex ID and adjust the multiplex ID in use map.  If this is the
    * last scope to be closed for this DMO type, request the context to drop
    * the backing temp table from the database when the database session is
    * closed.
    * 
    * @throws  PersistenceException
    *          if any error occurs within a listener while processing a
    *          change event.
    */
   private void closeMultiplexScope()
   throws PersistenceException
   {
      // Do we need to copy our current records before cleaning up?
      if (outputDMO != null)
      {
         // TODO:  this copy is unconditional;  this will need to be modified
         // to account for stop condition processing.
         copyAllRows((Temporary) getDMOProxy(), outputDMO, false);
      }
      
      Class<?> dmoIface = getDMOInterface();
      Context local = context.get();
      Map<Class<?>, Map<Integer, Integer>> map = local.inUseMPIDs;
      
      // Get the map of multiplex IDs which are in use for this DMO type
      // within our context.
      Map<Integer, Integer> inUse = map.get(dmoIface);
      
      // Access our reference count from this map.
      Integer rc = inUse.get(multiplexID);
      int refCount = rc.intValue();
      
      // Decrement the reference count.  If 0, remove the key, else store the
      // new value back in the map.
      if (--refCount == 0)
      {
         inUse.remove(multiplexID);
      }
      else
      {
         inUse.put(multiplexID, refCount);
      }
      
      // If the in-use map is now empty, the backing table will need to be
      // dropped.  Also remove the inUse map from the master map and remove
      // this buffer's domain if it is the master buffer.
      if (inUse.isEmpty())
      {
         map.remove(dmoIface);
         
         try
         {
            removeRecords(null, null, true);
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
         finally
         {
            local.emptyTableFlags.remove(multiplexID);
            
            // Remove any maps of reversible/undoables pinned by this buffer.
            // The virtual table associated with these is gone.
            bufferManager.removeAllReversibles(this);
            
            // Drop the table.
            local.dropTable(getTable());
         }
      }
      // Otherwise, if the current multiplexID is no longer in use, remove
      // all records for this multiplex ID from the table, but leave the
      // table intact for the other buffers still using it.
      else if (refCount == 0)
      {
         try
         {
            removeRecords(null, null, true);
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
         finally
         {
            local.emptyTableFlags.remove(multiplexID);
            
            // Remove any maps of reversible/undoables pinned by this buffer.
            // The virtual table associated with these is gone.
            bufferManager.removeAllReversibles(this);
         }
      }
   }

   /*
   private void count()
   {
      Persistence persistence = getPersistence();
      StringBuilder buf = new StringBuilder("select count(*) from ");
      buf.append(getDMOImplementationName());
      String hql = buf.toString();
      try
      {
         List list = persistence.list(hql, null, null, 0, 0, true);
         Number count = (Number) list.get(0);
         LOG.info(getTable() + " contains " + count + " records.");
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(-1,
                                         "Error counting temp table records",
                                         exc);
      }
   }
   */
   
   /**
    * Remove or change the multiplex ID for all record rows from the backing 
    * table which meet the specified criteria.  Because temp table instances
    * are multiplexed across nested buffer scopes and across multiple buffers
    * which use the same backing DMO type, the backing table cannot be dropped
    * until after the last buffer scope referencing the backing table in the
    * current context has been closed.  This method is invoked for all buffer
    * scope closing events.  This is also the worker method for the {@link
    * #clear(Temporary)} and {@link #delete(String, Object[])} methods.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.  <code>null</code> to delete all
    *          records whose multiplex IDs match that of this buffer.
    * @param   args
    *          Query substitution parameters required by the where clause.
    *          <code>null</code> if there are no substitutions or no where
    *          clause.
    * @param   forceDelete
    *          Force the records to be deleted from the table, not just
    *          marking them with a negative multiplex ID.
    * 
    * @throws  PersistenceException
    *          if an error occurs performing the bulk delete operation.
    */
   private void removeRecords(String where,
                              Object[] args,
                              boolean forceDelete)
   throws PersistenceException
   {
      // if we are not within a transaction, or this temp-table is not 
      // undoable
      forceDelete = forceDelete                          ||
                    !TransactionManager.isTransaction()  ||
                    !isUndoable();

      boolean hasWhere = (where != null);
      Persistence persistence = getPersistence();
      String hql = null;
      
      StringBuilder buf = new StringBuilder();
      Integer delMPID = 0;
      
      // Create portion of where clause which restricts by multiplex ID.
      buf.append(MULTIPLEX_FIELD_NAME);
      buf.append(" = ");
      buf.append(multiplexID);
      String mpidWhere = buf.toString();
      buf.setLength(0);
      
      if (forceDelete) 
      {
         buf.append("delete from ");
      }
      else
      {
         buf.append("update ");
         Context local = context.get();
         Map<Class<?>, Map<Integer, Integer>> map = local.inUseMPIDs;
         Map<Integer, Integer> inUse = map.get(getDMOInterface());
         
         delMPID = local.getNextMPID(inUse);
      }
      
      String dmoImplName = getDMOImplementationName();
      buf.append(dmoImplName);
      
      if (!forceDelete)
      {
         // do not delete, just change the multiplex ID
         buf.append(" set ");
         buf.append(MULTIPLEX_FIELD_NAME);
         buf.append(" = ");
         buf.append(delMPID);
      }

      buf.append(" where ");
      
      // Only prepend a test for multiplex ID if temp table is multiplexed or
      // if we have a where clause.  Otherwise, it is more efficient to omit
      // this test.
      if (hasWhere)
      {
         buf.append("(");
      }
      buf.append(mpidWhere);
      if (hasWhere)
      {
         buf.append(") and (");
      }
      
      Type[] types = null;
      if (hasWhere)
      {
         buf.append(HQLPreprocessor.get(where, this, args, false).getHQL());
         buf.append(")");
         if (args != null)
         {
            types = DBUtils.makeTypeArray(args);
         }
      }
      hql = buf.toString();
      
      int rowCount = persistence.deleteOrUpdate(hql, args, types);
      
      if (rowCount > 0)
      {
         changeBroker.forcePendingFlush(this);
         
         if (!forceDelete)
         {
            // if the records had their multiplex ID changed, register a
            // ReversibleBulkDelete for this where clause and arguments
            setReversibleBulkDelete(delMPID, hasWhere);
         }

         if (!isUndoable() && TransactionManager.isTransaction())
         {
            // force the records to be deleted, on db tx rollback
            ReversibleBulkDelete rev = new ReversibleBulkDelete(where, 
                                                                args, 
                                                                types, 
                                                                multiplexID);
            rev.setForceDelete(true);
            bufferManager.addReversible(this, null, rev, false);
         }
         
         boolean full = !hasWhere;
         
         if (!full)
         {
            // Execute a simple query to determine if there is any record in
            // this virtual table.
            buf.setLength(0);
            buf.append("select tt.id from ");
            buf.append(dmoImplName);
            buf.append(" tt where ");
            buf.append(mpidWhere);
            hql = buf.toString();
            
            List list = persistence.list(hql, null, null, 1, 0, true);
            full = (list == null);
         }
         
         if (full)
         {
            Context local = context.get();
            local.emptyTableFlags.put(multiplexID, Boolean.TRUE);
         }
         
         // Notify all interested parties that a bulk delete occurred.
         // Listening buffers may need to release deleted records.
         RecordChangeEvent.Type eventType =
            (full
             ? RecordChangeEvent.Type.FULL_BULK_DELETE
             : RecordChangeEvent.Type.RESTRICTED_BULK_DELETE);
         changeBroker.stateChanged(eventType, this, null, null, null);
      }
   }
   
   /**
    * Concrete implementation of <code>Reversible</code> which on rollback
    * reverses a bulk record delete action by setting the multiplex ID for 
    * the targeted DMO's back to its original value.
    */
   private final class ReversibleBulkDelete
   extends AbstractBulkReversible
   {      
      /** The multiplex ID used to mark the records as deleted. */
      private final int delMPID;

      /** 
       * If not-null, it represents the record which we want to rollback or
       * delete, only if it satisfies the where clause.
       */
      private final Serializable key;
      
      /** 
        * The where clause representing the records to be deleted. 
        * May be null.
        */
      private String where;

      /** The arguments for the where clause. May be null.*/
      private Object[] args;

      /** The associated type for each argument. May be null.*/
      private Type[] types;
      
      /** 
        * When <code>true</code>, this reversible will force the deletion of
        * the records, instead of updating the multiplex ID.
        */
      private boolean forceDelete = false;
      
      /**
       * Flag indicating if this bulk delete had a where clause or not.
       */
      private boolean hasWhere = false;
      
      /**
       * Constructor.
       *
       * @param   delMPID
       *          The multiplex ID value used to mark those records as deleted
       * @param   hasWhere
       *          Flag indicating if this bulk delete had a where clause.
       */
      ReversibleBulkDelete(int delMPID, boolean hasWhere)
      {
         this(delMPID, null);
         this.hasWhere = hasWhere;
      }

      /**
       * This constructor will limit the bulk delete to the where clause, if 
       * if specified. The c'tor will be used for re-roll forward actions, 
       * after a database tx rollback, for no-undo temp-tables. In this case,
       * the <code>delMPID</code> will specify the actual multiplex ID for
       * the records which should be deleted.
       *
       * @param   where  
       *          The conditions which targets the deleted records. May be 
       *          null.
       * @param   args
       *          The needed arguments for the <code>where</code> clause. May 
       *          be null.
       * @param   types
       *          The associated type for each argument. May be null.
       * @param   delMPID
       *          The multiplex ID value used to mark those records as deleted 
       */
      ReversibleBulkDelete(String where, 
                           Object[] args, 
                           Type[] types, 
                           int delMPID)
      {
         this(delMPID, null);
         this.where = where;
         this.args = args;
         this.types = types;
      }

      /**
       * Constructor to rollback a record which may deleted during a bulk
       * action.
       *
       * @param   delMPID
       *          the multiplex ID value used to mark those records as deleted
       * @param   key
       *          The key for the record which we want to delete or rollback.
       */
      ReversibleBulkDelete(int delMPID, Serializable key)
      {
         this.delMPID = delMPID;
         this.key = key;
      }
      
      /**
       * Get a string representation of the internal state of this object,
       * primarily for debugging purposes.
       *
       * @return  String representing this object's state.
       */
      public String toString()
      {
         StringBuilder buf = new StringBuilder("Bulk deletion");
         buf.append(" of ");
         buf.append(getDMOName());
         buf.append(" [MPID ");
         buf.append(delMPID);
         buf.append("]");
         
         return buf.toString();
      }
      
      /**
       * This method checks if this bulk reversible has deleted the record 
       * withthe given key.
       * 
       * @param   id
       *          Primary key of the record to be checked.
       *          
       * @return  <code>true</code> if the record was deleted by this bulk
       *          reversible.
       */
      protected boolean containsRecord(Serializable id)
      {
         if (!hasWhere)
         {
            // no where clause, the bulk delete affects the entire virtual
            // table
            return true;
         }

         Persistence persistence = getPersistence();
         if (bulkDeleteCheckHQL == null)
         {
            String entity = getEntityName();
            StringBuilder buf = new StringBuilder();
            buf.append("select tt.id from ");
            buf.append(entity);
            buf.append(" tt where tt.id=? and tt.");
            buf.append(MULTIPLEX_FIELD_NAME);
            buf.append("=?");
            bulkDeleteCheckHQL = buf.toString();
         }
         
         try
         {
            List list = persistence.list(bulkDeleteCheckHQL,
                                         new Object[] { id, delMPID },
                                         BULK_DELETE_CHECK_QUERY_PARM_TYPES,
                                         1,
                                         0,
                                         true);
            
            return (list != null);
         }
         catch (PersistenceException exc)
         {
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Error checking for record related to bulk delete",
                         exc);
            }
         }
         
         return false;
      }
      
      /**
       * Indicate whether this <code>Reversible</code> implementation stores its
       * DMO locally, as opposed to storing an identifier and having to retrieve
       * the DMO from the backing database, or not making a DMO available at all.
       * 
       * @return  <code>true</code> if the DMO is stored locally, else
       *          <code>false</code>.
       */
      protected boolean storesDMO()
      {
         return false;
      }
      
      /**
       * Get the unqualified interface name of the DMO being managed by this
       * reversible.
       * 
       * @return  DMO interface name.
       */
      protected String getDMOName()
      {
         return TemporaryBuffer.this.getDMOName();
      }
      
      /**
       * Set the <code>forceDelete</code> flag: when <code>true</code>, the 
       * records which satisfy the <code>where</code> clause (if any) will
       * be deleted.
       * 
       * @param   forceDelete
       */
      void setForceDelete(boolean forceDelete)
      {
         this.forceDelete = forceDelete;
      }
      
      /** 
       * No-op.
       * 
       * @return  Nothing.
       *
       * @throws  UnsupportedOperationException  
       *          No DMO is required for a bulk delete. 
       */
      protected Serializable getId()
      {
         throw new UnsupportedOperationException(
            "ReversibleBulkDelete does not require a DMO for persistable " +
            getDMOName());         
      }
      
      /**
       * Duplicate this reversible action. 
       * 
       * @param   key
       *          The key to set for the new reversible action.
       */
      Reversible duplicate(Serializable key)
      {
         ReversibleBulkDelete rev = new ReversibleBulkDelete(this.delMPID, 
                                                             key);
         rev.forceDelete = this.forceDelete;
         // copy the where clause only if this reversible will force the
         // delete and not just restore the multiplex ID. in the case of a
         // forced delete, the multiplex ID wasn't changed, so the where
         // clause is needed to ensure that the record with the specified key
         // can be deleted by bulk delete.
         if (forceDelete)
         {
            rev.where = this.where;
            rev.types = this.types;
            rev.args  = this.args;
         }

         return rev;
      }
            
      /** 
       * No-op.
       * 
       * @return  Nothing.
       * 
       * @throws  UnsupportedOperationException  
       *          always;  no DMO is required for a bulk delete. 
       */
      protected Persistable getDMO()
      {
         throw new UnsupportedOperationException(
            "ReversibleBulkDelete does not require a DMO for persistable " +
            getDMOName());         
      }

      /**
       * Set the underlying DMO for this reversible action.  This
       * implementation is a no-op.
       * 
       * @param   dmo
       *          The underlying DMO.
       */
      protected void setDMO(Persistable dmo)
      {
      }
      
      /**
       * Reverse the bulk deletion of the records which meet the specified 
       * criteria, by setting back their multiplex ID to a positive value.
       * <p>
       * If <code>forceDelete</code> is set to <code>true</code>, this method
       * will actually force the deletion of the records which satisfy the
       * <code>where</code> clause (if any) and have the specified multiplex
       * ID.
       *
       * @param   rollForward
       *          If <code>true</code>, this method is being invoked in order
       *          to re-roll forward a non-undoable change, after a
       *          database-level rollback.  This indicates the modification is
       *          made within a new transaction invoked after the
       *          database-level rollback.
       *          If <code>false</code>, this method is being invoked in order
       *          to roll back an undoable change within the context of an
       *          application-level, full transaction or sub-transaction
       *          rollback.
       * @param   transaction
       *          <code>true</code> if this rollback occurs at a full
       *          transaction boundary, else <code>false</code>.
       *
       * @throws  PersistenceException
       *          if an error occurs at the database or in the persistence
       *          layer when updating the records.
       */
      protected void rollback(boolean rollForward, boolean transaction)
      throws PersistenceException
      {
         if (forceDelete)
         {
            forceDelete();
            
            return;
         }
         
         StringBuilder buf = new StringBuilder("update ");         
         buf.append(getDMOImplementationName());
         buf.append(" set ");
         buf.append(MULTIPLEX_FIELD_NAME);
         buf.append(" = ");
         buf.append(multiplexID.intValue());
         
         executeQuery(buf);
      }
      
      /**
       * Execute an update or delete statement.
       * 
       * @param   buf
       *          The string used to construct the statement.
       *          
       * @throws  PersistenceException
       *          if an error occurs at the database or in the persistence
       *          layer when updating or deleting the records.
       */
      private void executeQuery(StringBuilder buf)
      throws PersistenceException
      {
         Persistence persistence = getPersistence();
         String hql = null;

         buf.append(" where (");
         buf.append(MULTIPLEX_FIELD_NAME);
         buf.append(" = ");
         buf.append(delMPID);
         buf.append(")");

         if (key != null)
         {
            buf.append(" and (id = ");
            buf.append(key);
            buf.append(")");
         }
         
         if (where != null)
         {
            buf.append(" and (");
            buf.append(where);
            buf.append(")");
         }
         
         hql = buf.toString();
         
         if (persistence.deleteOrUpdate(hql, args, types) > 0)
         {
            changeBroker.forcePendingFlush(TemporaryBuffer.this);

            // we need to set the empty table flag back to OFF, as this bulk
            // delete was rolled back and table is no longer empty
            Context local = context.get();
            local.emptyTableFlags.put(multiplexID, Boolean.FALSE);
         }         
      }
      
      /**
       * Remove the marked records from the table.  
       *
       * @throws  PersistenceException
       *          if an error occurs at the database or in the persistence
       *          layer when deleting the records.
       */
      private void forceDelete() 
      throws PersistenceException
      {        
         StringBuilder buf = new StringBuilder("delete from ");         
         buf.append(getDMOImplementationName());
         executeQuery(buf);
      }
   }

   /**
    * Helper which tracks the multiplex IDs currently in use by DMO interface
    * in the current client context, as well as the reference count of temp
    * tables created in the database using the context's current connection,
    * and domains of master/slave buffers.
    * <p>
    * This object also tracks tables which have been created during the
    * current database session, and drops them when the session is being
    * closed.  This is done as an optimization to prevent potential thrashing
    *in the event a class which creates a temporary buffer and then allows it
    * to go out of scope is invoked repeatedly in a loop.
    */
   private static class Context
   implements SessionListener,
              Commitable
   {
      /** Map of DMO interfaces to counts of multiplex IDs currently in use */
      private final Map<Class<?>, Map<Integer, Integer>> inUseMPIDs =
         new HashMap<Class<?>, Map<Integer, Integer>>();
      
      /** Map of open tables to most recent primary key ID generated */
      private final Map<Class<?>, Long> openTables =
         new HashMap<Class<?>, Long>();
      
      /** Primary keys which will be recycled after the next commit */
      private final Map<Class<?>, List<Serializable>> pendingReclaimedKeys =
         new HashMap<Class<?>, List<Serializable>>();
      
      /** Recycled primary keys available for re-use (lowest first) */
      private final Map<Class<?>, SortedSet<Serializable>> reclaimedKeys =
         new HashMap<Class<?>, SortedSet<Serializable>>();
      
      /** Names of active temp tables in the order they were created */
      private final Set<String> activeTables = new LinkedHashSet<String>();
      
      /** Map of flags indicating whether each virtual table is empty */
      private final Map<Integer, Boolean> emptyTableFlags =
         new HashMap<Integer, Boolean>();
      
      /** Singleton map of a unique key to session ID string */
      private final Map<String, String> sweepMap;
      
      /** Persistence service object for temp tables */
      private Persistence persistence;
      
      /** Map of master buffers to weak hash maps with slave buffer keys */
      private Map<TemporaryBuffer, Map<TemporaryBuffer, Object>> domains =
         new WeakHashMap<TemporaryBuffer, Map<TemporaryBuffer, Object>>();
      
      /** Next value to try when assigning a multiplex ID */
      private int nextMPID = 1;
      
      /**
       * Default constructor.
       */
      Context()
      {
         SecurityManager secMgr = SecurityManager.getInstance();
         String sid = String.valueOf(secMgr.getSessionId());
         sweepMap =
            Collections.singletonMap(TempTableHelper.UNIQUE_SUFFIX_KEY, sid);
      }
      
      /**
       * Invoked when a Hibernate session is about to close, but is still
       * available for cleanup processing.  This implementation drops all
       * temporary tables which have been created during the current database
       * session if the session is not being closed due to an error.
       * Otherwise, it just sets the temp table reference count to zero, so
       * that the connection provider is allowed to release the connection.
       *
       * @param   event
       *          Session event type.
       *
       * @return  <code>false</code> indicating that this session listener
       *          should remain registered after this method returns.
       *
       * @throws  PersistenceException
       *          if any error occurs dropping temp tables from the closing
       *          session.
       */
      public boolean sessionEvent(SessionListener.Event event)
      throws PersistenceException
      {
         switch (event)
         {
            case SESSION_CLOSING_NORMALLY:
               dropAllTables();
               break;
               
            case SESSION_CLOSING_WITH_ERROR:
               
               // On error, we want TempTableConnectionProvider to release the
               // broken connection, so make sure we are not pinning it open.
               activeTables.clear();
               
               // Clear remainder of state.
               Iterator<Class<?>> iter = openTables.keySet().iterator();
               while (iter.hasNext())
               {
                  Class<?> next = iter.next();
                  inUseMPIDs.remove(next);
                  iter.remove();
               }
               
               break;
               
            default:
               return false;
         }
         
         // Only invoked in SESSION_CLOSING_* cases, due to return in default
         // case in switch above.
         pendingReclaimedKeys.clear();
         reclaimedKeys.clear();
         
         return false;
      }
      
      /**
       * Invoked when this session listener is removed from the list of
       * registered session listeners.
       * <p>
       * This implementation does nothing.
       */
      public void deregisteredSessionListener()
      {
      }
      
      /**
       * On a full transaction commit, reclaim all of the primary keys which
       * have been surrendered by record deletes and clear the set of pending,
       * reclaimable keys.
       * 
       * @param   transaction
       *          <code>true</code> if called on a full transaction boundary;
       *          <code>false</code> if called on a subtransaction boundary.
       */
      public void commit(boolean transaction)
      {
         if (!transaction)
         {
            return;
         }
         
         for (Map.Entry<Class<?>, List<Serializable>> next
              : pendingReclaimedKeys.entrySet())
         {
            Class<?> dmoIface = next.getKey();
            List<Serializable> newIDs = next.getValue();
            SortedSet<Serializable> allIDs = reclaimedKeys.get(dmoIface);
            if (allIDs == null)
            {
               allIDs = new TreeSet<Serializable>();
               reclaimedKeys.put(dmoIface, allIDs);
            }
            allIDs.addAll(newIDs);
         }
         
         pendingReclaimedKeys.clear();
      }
      
      /**
       * On a full transaction rollback,clear the set of pending, reclaimable
       * keys.
       * 
       * @param   transaction
       *          <code>true</code> if called on a full transaction boundary;
       *          <code>false</code> if called on a subtransaction boundary.
       */
      public void rollback(boolean transaction)
      {
         if (transaction)
         {
            pendingReclaimedKeys.clear();
         }
      }
      
      /**
       * No-op.
       * 
       * @see com.goldencode.p2j.util.Commitable#rollbackPending()
       */
      public void rollbackPending()
      {
      }
      
      /**
       * No-op
       * 
       * @see com.goldencode.p2j.util.Commitable#validate(boolean)
       */
      public void validate(boolean transaction)
      {
      }
      
      /**
       * Initialize this object by storing the persistence service object and
       * registering as a session listener.  This method is tolerant of being
       * invoked multiple times, but only does its initialization work the
       * first time it is invoked for the current <code>Context</code>
       * instance.
       */
      void init()
      {
         if (persistence != null)
         {
            return;
         }
         
         persistence = PersistenceFactory.getInstance(
            DatabaseManager.TEMP_TABLE_DB);
         persistence.registerSessionListener(this,
                                             SessionListener.Scope.NONE);
      }
      
      /**
       * Report the number of temp tables which currently are in use in the
       * current context for the database associated with this buffer.
       *
       * @return  Number of temp tables associated with
       *          <code>persistence</code>.
       */
      int getTableCount()
      {
         return activeTables.size();
      }

      /**
       * Returns a context-local multiplex ID value.
       * 
       * @param   inUse
       *          Map of multiplex IDs currently in use in this context, to
       *          reference count of buffers using them.
       * 
       * @return  The next available multiplex ID.
       */
      Integer getNextMPID(Map<Integer, Integer> inUse)
      {
         Integer id = null;
         do
         {
            id = new Integer(nextMPID++);
         }
         while (inUse.containsKey(id));
         
         return id;
      }
      
      /**
       * Create a temp table in the database for the given DMO interface, if
       * it does not already exist.  This will also create all indexes defined
       * for the temp table.
       * 
       * @param   dmoIface
       *          DMO interface for which a backing temp table is needed.
       * @param   table
       *          Name of temp table to be created.
       */
      void createTable(Class<?> dmoIface, String table)
      {
         init();
         
         activeTables.add(table);
         if (!openTables.containsKey(dmoIface))
         {
            openTables.put(dmoIface, 0L);
            doCreateTable(dmoIface);
         }
      }
      
      /**
       * Decrement the reference count of active temp tables in use within
       * the current context.  This method does not actually drop a table,
       * since all table drops are deferred until the database session
       * actually closes.  This prevents potential thrashing in the event
       * a class which creates a temporary buffer and then allows it to go
       * out of scope is invoked repeatedly in a loop.
       * 
       * @param   table
       *          Name of temp table to be dropped.
       */
      void dropTable(String table)
      {
         activeTables.remove(table);
      }
      
      /**
       * Returns the next available primary key for the temp table associated
       * with the given DMO interface.
       *
       * @param   dmoIface
       *          Data model object interface.
       *
       * @return  The primary key for a new DMO associated with this buffer.
       * 
       * @throws  PersistenceException
       *          if <code>dmoIface</code> is unknown.
       */
      Serializable nextPrimaryKey(Class<?> dmoIface)
      throws PersistenceException
      {
         Long id = openTables.get(dmoIface);
         if (id == null)
         {
            throw new PersistenceException(
               "No temp table registered for DMO " + dmoIface.getName());
         }
         
         // If we have a recycled key, use it first.  Otherwise, allocate a
         // new key.
         SortedSet<Serializable> keys = reclaimedKeys.get(dmoIface);
         if (keys != null)
         {
            id = (Long) keys.first();
            keys.remove(id);
            if (keys.isEmpty())
            {
               reclaimedKeys.remove(dmoIface);
            }
         }
         else
         {
            id++;
            openTables.put(dmoIface, id);
         }
         
         return id;
      }
      
      /**
       * Prepare a single, primary key for reclamation.  On the next, full
       * transaction commit, this key will be made available for re-use.
       * Until then, it will be stored in a set of keys which are pending
       * reclamation.  This set will be cleared if the current transaction is
       * rolled back.
       * 
       * @param   dmoIface
       *          DMO interface for temporary table records.
       * @param   id
       *          Primary key to be recycled.
       */
      void reclaimKey(Class<?> dmoIface, Serializable id)
      {
         List<Serializable> keys = pendingReclaimedKeys.get(dmoIface);
         if (keys == null)
         {
            keys = new ArrayList<Serializable>();
            pendingReclaimedKeys.put(dmoIface, keys);
         }
         keys.add(id);
      }
      
      /**
       * Drop all temporary tables which have been created during the current
       * database session.  This typically occurs when the session is being
       * closed.
       *
       * @throws  PersistenceException
       *          if a database error occurs dropping any table or creating
       *          the transaction which contains these actions.
       */
      private void dropAllTables()
      throws PersistenceException
      {
         try
         {
            Iterator<Class<?>> iter = openTables.keySet().iterator();
            if (!iter.hasNext())
            {
               return;
            }
            
            boolean shouldCommit = persistence.beginTransaction();
            while (iter.hasNext())
            {
               Class<?> dmoIface = iter.next();
               doDropTable(dmoIface);
               iter.remove();
            }
            
            if (shouldCommit)
            {
               persistence.commit();
            }
         }
         finally
         {
            openTables.clear();
         }
      }
      
      /**
       * Issue the SQL statement(s) against the backing database necessary to
       * create the backing temp table needed for the given DMO interface, as
       * well as all indexes defined for that table.  Because temp table
       * instances are multiplexed across nested buffer scopes and across
       * multiple buffers which use the same backing DMO type, this method is
       * invoked only the first time a backing table is actually needed for a
       * particular DMO interface type in the current context.  Thereafter,
       * this method is not invoked again, unless the backing table has first
       * been dropped.
       *
       * @param   dmoIface
       *          DMO interface for which a backing temp table is needed.
       *
       * @throws  ErrorConditionException
       *          if an error occurs creating the table and silent error mode
       *          is disabled.
       *
       * @see     #doDropTable
       */
      private void doCreateTable(Class<?> dmoIface)
      {
         String sql = null;
         
         try
         {
            boolean shouldCommit = persistence.beginTransaction();
            Database database = persistence.getDatabase();
            
            // Create table(s).
            Iterator<String> tableIter =
               TempTableHelper.sqlTempTableCreate(database, dmoIface);
            while (tableIter.hasNext())
            {
               sql = tableIter.next();
               persistence.executeSQL(sql);
            }
            
            // Create indexes (if any).
            Iterator<String> indexIter =
               TempTableHelper.sqlTempTableIndexCreate(database, dmoIface);
            while (indexIter.hasNext())
            {
               sql = indexIter.next();
               sql = StringHelper.sweep(sql, sweepMap);
               persistence.executeSQL(sql);
            }
            
            if (shouldCommit)
            {
               persistence.commit();
            }
         }
         catch (PersistenceException exc)
         {
            openTables.remove(dmoIface);
            
            ErrorManager.recordOrThrowError(
               -1,
               "Error creating temp table:  " + getTableName(dmoIface) +
               " [" + sql + "]",
               exc);
         }
      }
      
      /**
       * Issue the SQL statement(s) against the backing database necessary to
       * drop the backing temp table for the given DMO interface.  Because
       * temp table instances are multiplexed across nested buffer scopes and
       * across multiple buffers which use the same backing DMO type, this
       * method is invoked only when the current database session is being
       * closed, after the last buffer scope referencing the backing table in
       * the current context has been closed.
       *
       * @param   dmoIface
       *          DMO interface for which the backing temp table should be
       *          dropped.
       *
       * @throws  PersistenceException
       *          if an error occurs dropping the table and silent error mode
       *          is disabled.
       *
       * @see     #doCreateTable
       */
      private void doDropTable(Class<?> dmoIface)
      throws PersistenceException
      {
         try
         {
            Database database = persistence.getDatabase();
            
            // Drop indexes (if any).
            Iterator<String> indexIter =
               TempTableHelper.sqlTempTableIndexDrop(database, dmoIface);
            while (indexIter.hasNext())
            {
               String sql = indexIter.next();
               sql = StringHelper.sweep(sql, sweepMap);
               persistence.executeSQL(sql);
            }
            
            // Drop table.
            Iterator<String> tableIter =
               TempTableHelper.sqlTempTableDrop(database, dmoIface);
            while (tableIter.hasNext())
            {
               String sql = tableIter.next();
               persistence.executeSQL(sql);
            }
         }
         catch (PersistenceException exc)
         {
            throw new PersistenceException(
               "Error dropping temp table:  " + getTableName(dmoIface),
               exc);
         }
      }
      
      /**
       * Get the backing temp table name associated with the given DMO
       * interface, primarily for debugging and error reporting purposes.
       *
       * @param   dmoIface
       *          DMO interface for which a table name is needed.
       *
       * @return  Name of the temp table associated with the given interface.
       */
      private String getTableName(Class<?> dmoIface)
      {
         Database tempDB = DatabaseManager.TEMP_TABLE_DB;
         String schema = DatabaseManager.getSchema(tempDB);
         Class<?> cls = DMOIndex.getImplementingClass(schema, dmoIface);
         
         return DatabaseManager.getBackingTableName(tempDB, cls);
      }
   }
   
   /**
    * An extended <code>ValidationHelper</code> which overrides methods which
    * must behave differently when validating temp tables.
    */
   protected final class TempValidationHelper
   extends ValidationHelper
   {
      /**
       * Override the parent's implementation to provide an empty list when
       * non-unique, dirty indexes are requested.  These are used only for
       * dirty share processing, which is not used for temp tables.
       * 
       * @param   entity
       *          DMO type name.
       * @param   unique
       *          If <code>true</code>, only unique indexes are checked.  If
       *          <code>false</code>, no indexes are checked.
       * @param   fullMatch
       *          If <code>true</code> every property in an index must be
       *          dirty for that index to be considered dirty.  If
       *          <code>false</code>, any dirty property in an index is enough
       *          to consider that index dirty.
       * 
       * @return  A list of names of indexes considered dirty, in ascending
       *          alphabetical order.
       */
      protected List<String> listDirtyIndexes(String entity,
                                              boolean unique,
                                              boolean fullMatch)
      {
         if (unique)
         {
            return super.listDirtyIndexes(entity, unique, fullMatch);
         }
         
         return Collections.emptyList();
      }
   }
   
   /**
    * An object registered with the <code>TransactionManager</code> to be
    * notified when its enclosing buffer's scope is closing.  This drives
    * cleanup of the backing temp table, which can take one of two forms:
    * <ul>
    *   <li>if no other buffers are using the backing table any longer, the
    *       table is dropped (along with all records);
    *   <li>otherwise, only those records associated with this buffer
    *       instance (as defined by this buffer's multiplex ID) are removed
    *       from the table, but the table itself is left intact.
    * </ul>
    */
   private class Multiplexer
   implements Finalizable
   {
      /**
       * No-op.
       */
      public void iterate()
      {
      }
      
      /**
       * No-op.
       */
      public void retry()
      {
      }
      
      /**
       * Cleans up the backing temp table by either dropping it (if no other
       * buffers are using it any longer), or by removing this buffer's
       * records from it.  If we are not already within a database-level
       * transaction, begin a transaction before making the changes, and
       * commit it afterward.
       */
      public void finished()
      {
         // Only allow cleanup if the outermost scope opened on this buffer
         // has been closed.
         if (getOpenScopeCount() > 0)
         {
            return;
         }
         
         try
         {
            Persistence persistence = getPersistence();
            boolean shouldCommit = persistence.beginTransaction();
            
            closeMultiplexScope();
            
            if (shouldCommit)
            {
               persistence.commit();
            }
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
      }
   }


   /**
    * Create a temporary table dynamically at runtime and assign it to a 
    * handle. The table is empty and will be populated with fileds, indexes 
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    * 
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    * @param   widgetPool
    *          The name of the widget pool that contains the new temp table, 
    *          if any, or empty string to create the object in the current 
    *          default widget pool.
    */
   public static void createTempTable(handle tth, String widgetPool)
   {
      // TODO Auto-generated method stub
   }

   /**
    * Create a temporary table dynamically at runtime and assign it to a 
    * handle. The table is empty and will be populated with fileds, indexes 
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    * 
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    * @param   widgetPool
    *          The name of the widget pool that contains the new temp table, 
    *          if any, or empty string to create the object in the current 
    *          default widget pool.
    */
   public static void createTempTable(handle tth, character widgetPool)
   {
      if (widgetPool.isUnknown())
      {
         TemporaryBuffer.createTempTable(tth, "");
      }
      else
      {
         TemporaryBuffer.createTempTable(tth, widgetPool.getValue());
      }
   }

   /**
    * Create a temporary table dynamically at runtime and assign it to a 
    * handle. The table is empty and will be populated with fileds, indexes 
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    *
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    */
   public static void createTempTable(handle tth)
   {
      TemporaryBuffer.createTempTable(tth, "");
   }

   /**
    * Duplicates the structure of an existing table, copying the field 
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handler assigned to a table buffer to create the copy.
    *
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical createTableLike(handle hbuf)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds a new field to the temp-table. This overloaded form creates field  
    * with properties se to their default values.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    *
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(String name, String type)
   {
      return addNewField(name, type, 0, "", "", "", "");
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    * @param   columnLabel
    *          The label of the column associated with the defined field
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(String name,
                              String type,
                              int    extent,
                              String format,
                              String initial,
                              String label,
                              String columnLabel)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    * @param   columnLabel
    *          The label of the column associated with the defined field
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character    name,
                              character    type,
                              int          extent,
                              character    format,
                              BaseDataType initial,
                              character    label,
                              character    columnLabel)
   {
      // TODO Auto-generated method stub
      return new logical();
   }


   /**
    * Adds a new index with the specified name to the temp-table. Newly 
    * created idex is empty and its components must be added individually 
    * with the <code>ADD-INDEX-FIELD()</code> method.
    * This form of the overloaded method creates an index that is not unique, 
    * not primary and not a word index.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to be created.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewIndex(String name)
   {
      return addNewIndex(name, false, false, false);
   }

   /**
    * Adds a new index with the specified name to the temp-table. Newly 
    * created idex is empty and its components must be added individually 
    * with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * @param   word
    *          TRUE if this is a word index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewIndex(String  name,
                              boolean unique,
                              boolean primary,
                              boolean word)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * @param   word
    *          TRUE if this is a word index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical addNewIndex(String  name,
                              logical unique,
                              logical primary,
                              logical word)
   {
      // TODO Auto-generated method stub
      return new logical();
   }


   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addFieldLike(String name, String source)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addFieldLike(character name, character source)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Handle to field name from which to copy the field.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addFieldLike(character name, handle source)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addIndexLike(String name, String sourceName, String dbTable)
   {
      // TODO Auto-generated method stub
      return new logical();
   }
   
   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addIndexLike(character name,
                               character sourceName,
                               character dbTable)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Signals that all the field and index definitions for a temp-table 
    * have been provided. After the call to this method no fields and indexes
    * can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that
    *          refer to this temp-table.
    *
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical tempTablePrepare(String name)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Signals that all the field and index definitions for a temp-table 
    * have been provided. After the call to this method no fields and indexes
    * can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that
    *          refer to this temp-table.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical tempTablePrepare(character name)
   {
      // TODO Auto-generated method stub
      return new logical();
   }


   /**
    * Obtain the handle to the buffer that was created by default for this 
    * temporary table. Every dynamic temp-table is created with at 
    * least one buffer. This buffer’s object handle is returned by this 
    * method. It cannot be called until the table has been prepared using 
    * <code>tempTablePrepare()</code> method.
    * This method is the P2J equivalent of <code>DEFAULT-BUFFER-HANDLE</code>
    * attribute.
    * 
    * @param   object (handle tth)
    * 
    * @return  The default buffer handle for the temporary table.
    */
   public handle getDefaultBufferHandle()
   {
      // TODO Auto-generated method stub
      return new handle();
   }

   /**
    * Obtain the handle to a particular field in the buffer.
    * This method is the P2J equivalent of <code>BUFFER-FIELD</code> method.
    * 
    * @param   name
    *          The field name.
    *
    * @return  A handle of the buffer of the requested field
    */
   public handle getField(String name)
   {
      // TODO Auto-generated method stub
      return new handle();
   }

   /**
    * Obtain the handle to a particular field in the buffer.
    * This method is the P2J equivalent of <code>BUFFER-FIELD</code> method.
    * 
    * @param   name
    *          The field name.
    *
    * @return  A handle of the buffer of the requested field
    */
   public handle getField(character name)
   {
      return name.isUnknown() ? new handle() : getField(name.getValue());
   }

   /**
    * Obtain the handle to a particular field in the buffer.
    * This method is the P2J equivalent of <code>BUFFER-FIELD</code> method.
    * 
    * @param   fieldNumber
    *          The number of a fields in the buffer.
    *
    * @return  A handle of the buffer of the requested field
    */
   public handle getField(int fieldNumber)
   {
      // TODO Auto-generated method stub
      return new handle();
   }

   /**
    * Obtain the handle to a particular field in the buffer.
    * This method is the P2J equivalent of <code>BUFFER-FIELD</code> method.
    * 
    * @param   name
    *          The number of a fields in the buffer.
    *
    * @return  A handle of the buffer of the requested field
    */
   public handle getField(NumberType fieldNo)
   {
      return fieldNo.isUnknown() ? new handle() : getField(fieldNo.intValue());
   }
}
